""" Paginatator methods for edX API implementations."""

from django.core.paginator import InvalidPage, Paginator
from django.http import Http404
from rest_framework import pagination
from rest_framework.response import Response


class DefaultPagination(pagination.PageNumberPagination):
    """
    Default paginator for APIs in edx-platform.

    This is configured in settings to be automatically used
    by any subclass of Django Rest Framework's generic API views.
    """
    page_size_query_param = "page_size"
    page_size = 10
    max_page_size = 100

    def get_paginated_response(self, data):
        """
        Annotate the response with pagination information.
        """
        return Response({
            'next': self.get_next_link(),
            'previous': self.get_previous_link(),
            'count': self.page.paginator.count,
            'num_pages': self.page.paginator.num_pages,
            'current_page': self.page.number,
            'start': (self.page.number - 1) * self.get_page_size(self.request),
            'results': data
        })


class NamespacedPageNumberPagination(pagination.PageNumberPagination):
    """
    Pagination scheme that returns results with pagination metadata
    embedded in a "pagination" attribute.  Can be used with data
    that comes as a list of items, or as a dict with a "results"
    attribute that contains a list of items.
    """

    page_size_query_param = "page_size"

    def get_result_count(self):
        """
        Returns total number of results
        """
        return self.page.paginator.count

    def get_num_pages(self):
        """
        Returns total number of pages the results are divided into
        """
        return self.page.paginator.num_pages

    def get_paginated_response(self, data):
        """
        Annotate the response with pagination information
        """
        metadata = {
            'next': self.get_next_link(),
            'previous': self.get_previous_link(),
            'count': self.get_result_count(),
            'num_pages': self.get_num_pages(),
        }
        if isinstance(data, dict):
            if 'results' not in data:
                raise TypeError('Malformed result dict')
            data['pagination'] = metadata
        else:
            data = {
                'results': data,
                'pagination': metadata,
            }
        return Response(data)


def paginate_search_results(object_class, search_results, page_size, page):
    """
    Takes search results and returns a Page object populated
    with db objects for that page.

    :param object_class: Model class to use when querying the db for objects.
    :param search_results: search results.
    :param page_size: Number of results per page.
    :param page: Page number.
    :return: Paginator object with model objects
    """
    paginator = Paginator(search_results['results'], page_size)

    # This code is taken from within the GenericAPIView#paginate_queryset method.
    # It is common code, but
    try:
        page_number = paginator.validate_number(page)
    except InvalidPage as page_error:
        if page == 'last':
            page_number = paginator.num_pages
        else:
            raise Http404("Page is not 'last', nor can it be converted to an int.") from page_error

    try:
        paged_results = paginator.page(page_number)
    except InvalidPage as exception:
        raise Http404(
            "Invalid page {page_number}: {message}".format(
                page_number=page_number,
                message=str(exception)
            )
        ) from exception

    search_queryset_pks = [item['data']['pk'] for item in paged_results.object_list]
    queryset = object_class.objects.filter(pk__in=search_queryset_pks)

    def ordered_objects(primary_key):
        """ Returns database object matching the search result object"""
        for obj in queryset:
            if obj.pk == primary_key:
                return obj
        return None

    # map over the search results and get a list of database objects in the same order
    object_results = list(map(ordered_objects, search_queryset_pks))
    paged_results.object_list = object_results

    return paged_results



""" JWT Authentication class. """

import logging

import jwt
from django.contrib.auth import get_user_model
from django.middleware.csrf import CsrfViewMiddleware
from edx_django_utils.monitoring import set_custom_attribute
from rest_framework import exceptions
from rest_framework_jwt.authentication import JSONWebTokenAuthentication

from edx_rest_framework_extensions.auth.jwt.constants import USE_JWT_COOKIE_HEADER
from edx_rest_framework_extensions.auth.jwt.decoder import configured_jwt_decode_handler
from edx_rest_framework_extensions.settings import get_setting


logger = logging.getLogger(__name__)


class CSRFCheck(CsrfViewMiddleware):
    def _reject(self, request, reason):
        # Return the failure reason instead of an HttpResponse
        return reason


class JwtAuthentication(JSONWebTokenAuthentication):
    """
    JSON Web Token based authentication.

    This authentication class is useful for authenticating a JWT using a secret key. Clients should authenticate by
    passing the token key in the "Authorization" HTTP header, prepended with the string `"JWT "`.

    This class relies on the JWT_AUTH being configured for the application as well as JWT_PAYLOAD_USER_ATTRIBUTES
    being configured in the EDX_DRF_EXTENSIONS config.

    At a minimum, the JWT payload must contain a username. If an email address
    is provided in the payload, it will be used to update the retrieved user's
    email address associated with that username.

    Example Header:
        Authorization: JWT eyJhbGciOiJSUzUxMiIsInR5cCI6IkpXVCJ9.eyJzdWIiOiJmYzJiNzIwMTE0YmIwN2I0NjVlODQzYTc0ZWM2ODNlNiIs
        ImFkbWluaXN0cmF0b3IiOmZhbHNlLCJuYW1lIjoiaG9ub3IiLCJleHA.QHDXdo8gDJ5p9uOErTLZtl2HK_61kgLs71VHp6sLx8rIqj2tt9yCfc_0
        JUZpIYMkEd38uf1vj-4HZkzeNBnZZZ3Kdvq7F8ZioREPKNyEVSm2mnzl1v49EthehN9kwfUgFgPXfUh-pCvLDqwCCTdAXMcTJ8qufzEPTYYY54lY
    """

    def get_jwt_claim_attribute_map(self):
        """ Returns a mapping of JWT claims to user model attributes.

        Returns
            dict
        """
        return get_setting('JWT_PAYLOAD_USER_ATTRIBUTE_MAPPING')

    def get_jwt_claim_mergeable_attributes(self):
        """ Returns a list of user model attributes that should be merged into from the JWT.

        Returns
            list
        """
        return get_setting('JWT_PAYLOAD_MERGEABLE_USER_ATTRIBUTES')

    def authenticate(self, request):
        try:
            user_and_auth = super().authenticate(request)

            # Unauthenticated, CSRF validation not required
            if not user_and_auth:
                return user_and_auth

            # Not using JWT cookies, CSRF validation not required
            use_jwt_cookie_requested = request.META.get(USE_JWT_COOKIE_HEADER)
            if not use_jwt_cookie_requested:
                return user_and_auth

            self.enforce_csrf(request)

            # CSRF passed validation with authenticated user
            return user_and_auth

        except jwt.InvalidTokenError as token_error:
            # Note: I think this case is not used, but will monitor the custom attribute to verify.
            set_custom_attribute('jwt_auth_failed', 'InvalidTokenError:{}'.format(repr(token_error)))
            raise exceptions.AuthenticationFailed() from token_error
        except Exception as exception:
            # Errors in production do not need to be logged (as they may be noisy),
            # but debug logging can help quickly resolve issues during development.
            logger.debug('Failed JWT Authentication,', exc_info=exception)
            # Note: I think this case should only include AuthenticationFailed and PermissionDenied,
            #   but will monitor the custom attribute to verify.
            set_custom_attribute('jwt_auth_failed', 'Exception:{}'.format(repr(exception)))
            raise

    def authenticate_credentials(self, payload):
        """Get or create an active user with the username contained in the payload."""
        # TODO it would be good to refactor this heavily-nested function.
        # pylint: disable=too-many-nested-blocks
        username = payload.get('preferred_username') or payload.get('username')
        if username is None:
            raise exceptions.AuthenticationFailed('JWT must include a preferred_username or username claim!')
        try:
            user, __ = get_user_model().objects.get_or_create(username=username)
            attributes_updated = False
            attribute_map = self.get_jwt_claim_attribute_map()
            attributes_to_merge = self.get_jwt_claim_mergeable_attributes()
            for claim, attr in attribute_map.items():
                payload_value = payload.get(claim)

                if attr in attributes_to_merge:
                    # Merge new values that aren't already set in the user dictionary
                    if not payload_value:
                        continue

                    current_value = getattr(user, attr, None)

                    if current_value:
                        for (key, value) in payload_value.items():
                            if key in current_value:
                                if current_value[key] != value:
                                    logger.info(
                                        'Updating attribute %s[%s] for user %s with value %s',
                                        attr,
                                        key,
                                        user.id,
                                        value,
                                    )
                                    current_value[key] = value
                                    attributes_updated = True
                            else:
                                logger.info(
                                    'Adding attribute %s[%s] for user %s with value %s',
                                    attr,
                                    key,
                                    user.id,
                                    value,
                                )
                                current_value[key] = value
                                attributes_updated = True
                    else:
                        logger.info('Updating attribute %s for user %s with value %s', attr, user.id, payload_value)
                        setattr(user, attr, payload_value)
                        attributes_updated = True
                else:
                    if getattr(user, attr) != payload_value and payload_value is not None:
                        logger.info('Updating attribute %s for user %s with value %s', attr, user.id, payload_value)
                        setattr(user, attr, payload_value)
                        attributes_updated = True

            if attributes_updated:
                user.save()
        except Exception as authentication_error:
            msg = f'[edx-drf-extensions] User retrieval failed for username {username}.'
            logger.exception(msg)
            raise exceptions.AuthenticationFailed(msg) from authentication_error

        return user

    def enforce_csrf(self, request):
        """
        Enforce CSRF validation for Jwt cookie authentication.

        Copied from SessionAuthentication.
        See https://github.com/encode/django-rest-framework/blob/3f19e66d9f2569895af6e91455e5cf53b8ce5640/rest_framework/authentication.py#L131-L141  # noqa E501 line too long
        """
        check = CSRFCheck()  # pylint: disable=no-value-for-parameter
        # populates request.META['CSRF_COOKIE'], which is used in process_view()
        check.process_request(request)
        reason = check.process_view(request, None, (), {})
        if reason:
            # CSRF failed, bail with explicit error message
            raise exceptions.PermissionDenied('CSRF Failed: %s' % reason)


def is_jwt_authenticated(request):
    successful_authenticator = getattr(request, 'successful_authenticator', None)
    if not isinstance(successful_authenticator, JSONWebTokenAuthentication):
        return False
    if not getattr(request, 'auth', None):
        logger.error(
            'Unexpected error: Used JwtAuthentication, '
            'but the request auth attribute was not populated with the JWT.'
        )
        return False
    return True


def get_decoded_jwt_from_auth(request):
    """
    Grab jwt from request.auth in request if possible.

    Returns a decoded jwt dict if it can be found.
    Returns None if the jwt is not found.
    """
    if not is_jwt_authenticated(request):
        return None

    return configured_jwt_decode_handler(request.auth)



""" Permission classes. """
import logging

from opaque_keys.edx.keys import CourseKey
from rest_framework.permissions import BasePermission, IsAuthenticated

from edx_rest_framework_extensions.auth.jwt.authentication import is_jwt_authenticated
from edx_rest_framework_extensions.auth.jwt.decoder import (
    decode_jwt_filters,
    decode_jwt_is_restricted,
    decode_jwt_scopes,
)


log = logging.getLogger(__name__)


class IsSuperuser(BasePermission):
    """ Allows access only to superusers. """

    def has_permission(self, request, view):
        return request.user and request.user.is_superuser


class IsStaff(BasePermission):
    """
    Allows access to "global" staff users..
    """
    def has_permission(self, request, view):
        return request.user.is_staff


class IsUserInUrl(BasePermission):
    """
    Allows access if the requesting user matches the user in the URL.
    """
    def has_permission(self, request, view):
        allowed = request.user.username.lower() == get_username_param(request)
        if not allowed:
            log.info("Permission IsUserInUrl: not satisfied for requesting user %s.", request.user.username)
        return allowed


class JwtRestrictedApplication(BasePermission):
    """
    Allows access if the request was successfully authenticated with JwtAuthentication
    by a RestrictedApplication.
    """
    message = 'Not a Restricted JWT Application.'

    def has_permission(self, request, view):
        ret_val = is_jwt_authenticated(request) and decode_jwt_is_restricted(request.auth)
        log.debug("Permission JwtRestrictedApplication: returns %s.", ret_val)
        return ret_val


class NotJwtRestrictedApplication(BasePermission):
    """
    Allows access if either the request was not authenticated with JwtAuthentication, or
    if it was successfully authenticated with JwtAuthentication and the Jwt was not
    flagged as restricted.

    Note: Anonymous access will also pass this permission.

    """
    def has_permission(self, request, view):
        return not JwtRestrictedApplication().has_permission(request, view)


class JwtHasScope(BasePermission):
    """
    The request is authenticated as a user and the token used has the right scope.
    """
    message = 'JWT missing required scopes.'

    def has_permission(self, request, view):
        jwt_scopes = decode_jwt_scopes(request.auth)
        required_scopes = set(getattr(view, 'required_scopes', []))
        allowed = bool(required_scopes) and required_scopes.issubset(jwt_scopes)
        if not allowed:
            log.warning(
                "Permission JwtHasScope: required scopes '%s' are not a subset of the token's scopes '%s'.",
                required_scopes,
                jwt_scopes,
            )
        return allowed


class JwtHasContentOrgFilterForRequestedCourse(BasePermission):
    """
    The JWT used to authenticate contains the appropriate content provider
    filter for the requested course resource.
    """
    message = 'JWT missing required content_org filter.'

    def has_permission(self, request, view):
        """
        Ensure that the course_id kwarg provided to the view contains one
        of the organizations specified in the content provider filters
        in the JWT used to authenticate.
        """
        course_key = CourseKey.from_string(view.kwargs.get('course_id'))
        jwt_filters = decode_jwt_filters(request.auth)
        for filter_type, filter_value in jwt_filters:
            if filter_type == 'content_org' and filter_value == course_key.org:
                return True
        log.warning(
            "Permission JwtHasContentOrgFilterForRequestedCourse: no filter found for %s.",
            course_key.org,
        )
        return False


class JwtHasUserFilterForRequestedUser(BasePermission):
    """
    The JWT used to authenticate contains the appropriate user filter for the
    requested user resource.
    """
    message = 'JWT missing required user filter.'

    def has_permission(self, request, view):
        """
        If the JWT has a user filter, verify that the filtered
        user value matches the user in the URL.
        """
        user_filter = self._get_user_filter(request)
        if not user_filter:
            # no user filters are present in the token to limit access
            return True

        username_param = get_username_param(request)
        allowed = user_filter == username_param
        if not allowed:
            log.warning(
                "Permission JwtHasUserFilterForRequestedUser: user_filter %s doesn't match username %s.",
                user_filter,
                username_param,
            )
        return allowed

    def _get_user_filter(self, request):
        jwt_filters = decode_jwt_filters(request.auth)
        for filter_type, filter_value in jwt_filters:
            if filter_type == 'user':
                if filter_value == 'me':
                    filter_value = request.user.username.lower()
                return filter_value
        return None


class LoginRedirectIfUnauthenticated(IsAuthenticated):
    """
    A DRF permission class that will login redirect unauthorized users.

    It can be used to convert a plain Django view that was using @login_required
    into a DRF APIView, which is useful to enable our DRF JwtAuthentication class.

    Requires JwtRedirectToLoginIfUnauthenticatedMiddleware to work.

    """


_NOT_JWT_RESTRICTED_PERMISSIONS = NotJwtRestrictedApplication & (IsStaff | IsUserInUrl)
_JWT_RESTRICTED_PERMISSIONS = (
    JwtRestrictedApplication &
    JwtHasScope &
    JwtHasContentOrgFilterForRequestedCourse &
    JwtHasUserFilterForRequestedUser
)
JWT_RESTRICTED_APPLICATION_OR_USER_ACCESS = (
    IsAuthenticated &
    (_NOT_JWT_RESTRICTED_PERMISSIONS | _JWT_RESTRICTED_PERMISSIONS)
)


def get_username_param(request):
    user_parameter_name = 'username'
    url_username = (
        getattr(request, 'parser_context', {}).get('kwargs', {}).get(user_parameter_name, '') or
        request.GET.get(user_parameter_name, '')
    )
    return url_username.lower()



""" Session Authentication classes. """

from rest_framework.authentication import SessionAuthentication


class SessionAuthenticationAllowInactiveUser(SessionAuthentication):
    """Ensure that the user is logged in, but do not require the account to be active.

    We use this in the special case that a user has created an account,
    but has not yet activated it.  We still want to allow the user to
    enroll in courses, so we remove the usual restriction
    on session authentication that requires an active account.

    You should use this authentication class ONLY for end-points that
    it's safe for an un-activated user to access.  For example,
    we can allow a user to update his/her own enrollments without
    activating an account.

    """
    def authenticate(self, request):
        """Authenticate the user, requiring a logged-in account and CSRF.

        This is exactly the same as the `SessionAuthentication` implementation,
        with the `user.is_active` check removed.

        Args:
            request (HttpRequest)

        Returns:
            Tuple of `(user, token)`

        Raises:
            PermissionDenied: The CSRF token check failed.

        """
        # Get the underlying HttpRequest object
        request = request._request  # pylint: disable=protected-access
        user = getattr(request, 'user', None)

        # Unauthenticated, CSRF validation not required
        # This is where regular `SessionAuthentication` checks that the user is active.
        # We have removed that check in this implementation.
        # But we added a check to prevent anonymous users since we require a logged-in account.
        if not user or user.is_anonymous:
            return None

        self.enforce_csrf(request)

        # CSRF passed with authenticated user
        return (user, None)



""" Tests for permission classes. """

from collections import namedtuple
from itertools import product
from unittest.mock import Mock, patch

import ddt
from django.contrib.auth.models import AnonymousUser
from django.test import RequestFactory, TestCase
from rest_framework.authentication import SessionAuthentication
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework_jwt.authentication import JSONWebTokenAuthentication

from edx_rest_framework_extensions import permissions
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from edx_rest_framework_extensions.auth.jwt.tests.utils import generate_jwt
from edx_rest_framework_extensions.tests import factories
from edx_rest_framework_extensions.tests.factories import UserFactory


@ddt.ddt
class IsSuperuserTests(TestCase):
    """ Tests for the IsSuperuser permission class. """

    @ddt.data(True, False)
    def test_superuser_has_permission(self, has_permission):
        request = RequestFactory().get('/')
        request.user = factories.UserFactory(is_superuser=has_permission)
        permission = permissions.IsSuperuser()
        self.assertEqual(permission.has_permission(request, None), has_permission)

    @ddt.data(None, AnonymousUser())
    def test_invalid_user_has_no_permission(self, user):
        request = RequestFactory().get('/')
        request.user = user
        permission = permissions.IsSuperuser()
        self.assertFalse(permission.has_permission(request, None))


@ddt.ddt
class IsStaffTests(TestCase):
    """ Tests for the IsStaff permission class. """

    @ddt.data(True, False)
    def test_has_permission(self, is_staff):
        request = RequestFactory().get('/')
        request.user = factories.UserFactory(is_staff=is_staff)
        self.assertEqual(permissions.IsStaff().has_permission(request, None), is_staff)


@ddt.ddt
class IsUserInUrlTests(TestCase):
    """ Tests for the IsUserInUrl permission class. """

    def _create_request(self, user, username_in_param=None, username_in_resource=None):
        url = '/'
        if username_in_param:
            url += f'?username={username_in_param}'
        request = RequestFactory().get(url)
        request.user = user
        if username_in_resource:
            request.parser_context = {
                "kwargs": {"username": username_in_resource}
            }
        return request

    @ddt.data(True, False)
    def test_user_in_url_param(self, has_user_in_url):
        user = factories.UserFactory(username='this_user')
        request = self._create_request(
            user,
            username_in_param=user.username if has_user_in_url else 'another_user',
        )
        self.assertEqual(permissions.IsUserInUrl().has_permission(request, None), has_user_in_url)

    @ddt.data(True, False)
    def test_user_in_url_resource(self, has_user_in_resource):
        user = factories.UserFactory(username='this_user')
        request = self._create_request(
            user,
            username_in_resource=user.username if has_user_in_resource else 'another_user',
        )
        self.assertEqual(permissions.IsUserInUrl().has_permission(request, None), has_user_in_resource)

    def test_resource_takes_precedence_over_param(self):
        user = factories.UserFactory(username='this_user')
        request = self._create_request(
            user,
            username_in_resource='another_user',
            username_in_param='this_user',
        )
        self.assertFalse(permissions.IsUserInUrl().has_permission(request, None))

        request = self._create_request(
            user,
            username_in_resource='this_user',
            username_in_param='another_user',
        )
        self.assertTrue(permissions.IsUserInUrl().has_permission(request, None))


@ddt.ddt
class JwtApplicationPermissionsTests(TestCase):
    """ Tests for the JwtRestrictedApplication and NotJwtRestrictedApplication permission classes. """

    @ddt.data(
        *product(
            (permissions.JwtRestrictedApplication, permissions.NotJwtRestrictedApplication),
            (JwtAuthentication, JSONWebTokenAuthentication, SessionAuthentication, None),
            (True, False),
        )
    )
    @ddt.unpack
    def test_has_permission(self, permission_class, authentication_class, is_restricted):
        request = RequestFactory().get('/')
        request.successful_authenticator = authentication_class() if authentication_class else None
        request.user = factories.UserFactory()
        request.auth = generate_jwt(request.user, is_restricted=is_restricted)

        is_jwt_auth_subclass = issubclass(type(request.successful_authenticator), JSONWebTokenAuthentication)

        has_permission = permission_class().has_permission(request, view=None)
        expected_restricted_permission = is_restricted and is_jwt_auth_subclass
        if permission_class == permissions.JwtRestrictedApplication:
            self.assertEqual(has_permission, expected_restricted_permission)
        else:
            self.assertEqual(has_permission, not expected_restricted_permission)


@ddt.ddt
class JwtHasScopeTests(TestCase):
    """ Tests for JwtHasScope permission class. """
    def setUp(self):
        super().setUp()
        self.user = UserFactory()

    @ddt.data(
        (JwtAuthentication(), ('test:read',), ('test:read',), True),  # match
        (JwtAuthentication(), ('test:write'), ('test:read',), False),  # mismatch
        (JwtAuthentication(), ('test:read'), (), False),  # empty on API
        (JwtAuthentication(), ('test:read'), None, False),  # missing on API
        (JwtAuthentication(), (), ('test:read',), False),  # missing on jwt
        (JwtAuthentication(), (), None, False),  # missing on both
        (None, (), ('test:read'), False),  # no auth
    )
    @ddt.unpack
    def test_has_permission(self, authentication_class, jwt_scopes, required_scopes, expected_result):
        request = RequestFactory().get('/')
        request.successful_authenticator = authentication_class
        request.auth = generate_jwt(self.user, scopes=jwt_scopes)
        if required_scopes is None:
            view = APIView()
        else:
            view = Mock(required_scopes=required_scopes)
        self.assertEqual(permissions.JwtHasScope().has_permission(request, view), expected_result)


@ddt.ddt
class JwtHasContentOrgFilterForRequestedCourseTests(TestCase):
    """ Tests for JwtHasContentOrgFilterForRequestedCourse permission class. """
    def setUp(self):
        super().setUp()
        self.user = UserFactory()

    @ddt.data(
        (JwtAuthentication(), ['content_org:edX'], {'course_id': 'course-v1:edX+DemoX+Demo_Course'}, True),
        (JwtAuthentication(), ['content_org:TestX'], {'course_id': 'course-v1:edX+DemoX+Demo_Course'}, False),
        (JwtAuthentication(), ['test:TestX'], {'course_id': 'course-v1:edX+DemoX+Demo_Course'}, False),
        (JwtAuthentication(), [], {'course_id': 'course-v1:edX+DemoX+Demo_Course'}, False),
        (None, [], {'course_id': 'course-v1:edX+DemoX+Demo_Course'}, False),
    )
    @ddt.unpack
    def test_has_permission(self, authentication_class, jwt_filters, view_kwargs, expected_result):
        request = RequestFactory().get('/')
        request.successful_authenticator = authentication_class
        request.auth = generate_jwt(self.user, filters=jwt_filters)
        view = Mock(kwargs=view_kwargs)
        self.assertEqual(
            permissions.JwtHasContentOrgFilterForRequestedCourse().has_permission(request, view),
            expected_result,
        )


@ddt.ddt
class JwtHasUserFilterForRequestedUserTests(TestCase):
    """ Tests for JwtHasUserFilterForRequestedUserTests permission class. """

    def _create_request(self, user_filters, requested_username):
        url = f'/?username={requested_username}'
        request = RequestFactory().get(url)
        request.user = UserFactory(username='this_user')
        request.auth = generate_jwt(request.user, filters=user_filters)
        return request

    @ddt.data(
        # no user filters
        ([], 'this_user', True),
        ([], 'another_user', True),

        # accessing self
        (['user:me'], 'this_user', True),
        (['user:this_user'], 'this_user', True),

        # accessing another
        (['user:me'], 'another_user', False),
        (['user:this_user'], 'another_user', False),
        (['user:another_user'], 'another_user', True),
    )
    @ddt.unpack
    def test_has_permission(self, user_filters, requested_username, expected_result):
        request = self._create_request(user_filters, requested_username)
        self.assertEqual(
            permissions.JwtHasUserFilterForRequestedUser().has_permission(request, None),
            expected_result,
        )


@ddt.ddt
class JwtRestrictedApplicationOrUserAccessTests(TestCase):
    ExpectedLog = namedtuple('ExpectedLog', 'method, text, param')

    IsUserInUrlLog = ExpectedLog('info', 'IsUserInUrl', None)
    JwtIsUnrestrictedDebugLog = ExpectedLog('debug', 'JwtRestrictedApplication', False)
    JwtIsRestrictedDebugLog = ExpectedLog('debug', 'JwtRestrictedApplication', True)
    JwtScopesErrorLog = ExpectedLog('warning', 'JwtHasScope', None)
    JwtOrgFilterErrorLog = ExpectedLog('warning', 'JwtHasContentOrgFilterForRequestedCourse', None)
    JwtUserFilterErrorLog = ExpectedLog('warning', 'JwtHasUserFilterForRequestedUser', None)

    class SomeClassView(APIView):
        authentication_classes = (JwtAuthentication, SessionAuthentication)
        permission_classes = (permissions.JWT_RESTRICTED_APPLICATION_OR_USER_ACCESS,)
        required_scopes = ['required_scope']

        def get(self, request, course_id=None):  # pylint: disable=unused-argument
            return Response(data="Success")

    def _create_user(self, is_staff=False):
        return UserFactory(username='this_user', is_staff=is_staff)

    def _create_request(self, username_in_url=None, auth_header=None):
        url = '/'
        if username_in_url:
            url += f'?username={username_in_url}'
        extra = {"HTTP_AUTHORIZATION": auth_header} if auth_header else {}
        return RequestFactory().get(url, **extra)

    def _create_session(self, request, user):
        request.user = user

    def _create_jwt_header(self, user, is_restricted=False, scopes=None, filters=None):
        token = generate_jwt(user, is_restricted=is_restricted, scopes=scopes, filters=filters)
        return f"JWT {token}"

    def _assert_log(self, mock_log, expected_log):
        mock_log_method = getattr(mock_log, expected_log.method)
        self.assertTrue(mock_log_method.called)
        self.assertIn(expected_log.text, mock_log_method.call_args_list[0][0][0])
        if expected_log.param is not None:
            self.assertEqual(mock_log_method.call_args_list[0][0][1], expected_log.param)

    def test_anonymous_fails(self):
        request = self._create_request()
        response = self.SomeClassView().dispatch(request)
        self.assertEqual(response.status_code, 401)

    def test_session_staff_succeeds(self):
        user = self._create_user(is_staff=True)
        request = self._create_request()
        self._create_session(request, user)

        response = self.SomeClassView().dispatch(request)
        self.assertEqual(response.status_code, 200)

    @patch('edx_rest_framework_extensions.permissions.log')
    def test_session_user_not_in_url_fails(self, mock_log):
        user = self._create_user()
        request = self._create_request()
        self._create_session(request, user)

        response = self.SomeClassView().dispatch(request)
        self.assertEqual(response.status_code, 403)
        self._assert_log(mock_log, self.IsUserInUrlLog)

    def test_session_user_in_url_succeeds(self):
        user = self._create_user()
        request = self._create_request(username_in_url=user.username)
        self._create_session(request, user)

        response = self.SomeClassView().dispatch(request)
        self.assertEqual(response.status_code, 200)

    @ddt.data(
        # unrestricted
        {
            "is_restricted": False,
            "is_user_in_url": True,
            "expected_response": 200,
            "expected_log": JwtIsUnrestrictedDebugLog
        },
        {
            "is_restricted": False,
            "is_user_in_url": False,
            "expected_response": 403,
            "expected_log": IsUserInUrlLog
        },

        # restricted (note: further test cases for scopes and filters are in tests below)
        {
            "is_restricted": True,
            "is_user_in_url": True,
            "expected_response": 403,
            "expected_log": JwtScopesErrorLog
        },
        {
            "is_restricted": True,
            "is_user_in_url": False,
            "expected_response": 403,
            "expected_log": JwtScopesErrorLog
        },
    )
    @ddt.unpack
    def test_jwt_without_scopes_and_filters(
            self,
            is_restricted,
            is_user_in_url,
            expected_response,
            expected_log
    ):
        with patch('edx_rest_framework_extensions.permissions.log') as mock_log:
            user = self._create_user()

            auth_header = self._create_jwt_header(user, is_restricted=is_restricted)
            request = self._create_request(
                username_in_url=user.username if is_user_in_url else None,
                auth_header=auth_header,
            )

            response = self.SomeClassView().dispatch(request)
            self.assertEqual(response.status_code, expected_response)
            self._assert_log(mock_log, expected_log)

    @ddt.data(
        # valid scopes
        {
            "scopes": ['required_scope'], "expected_response": 200,
            "expected_log": JwtIsRestrictedDebugLog
        },
        {
            "scopes": ['required_scope', 'another_scope'],
            "expected_response": 200, "expected_log": JwtIsRestrictedDebugLog
        },

        # invalid scopes
        {
            "scopes": [], "expected_response": 403,
            "expected_log": JwtScopesErrorLog
        },
        {
            "scopes": ['another_scope'], "expected_response": 403,
            "expected_log": JwtScopesErrorLog
        },
    )
    @ddt.unpack
    def test_jwt_scopes(self, scopes, expected_response, expected_log):
        self._assert_jwt_restricted_case(
            scopes=scopes,
            filters=['content_org:some_org'],
            is_user_in_url=False,
            expected_response=expected_response,
            expected_log=expected_log,
        )

    @ddt.data(
        # valid org filters
        {
            "filters": ['content_org:some_org', 'content_org:another_org'],
            "expected_response": 200,
            "expected_log": JwtIsRestrictedDebugLog
        },

        # invalid org filters
        {
            "filters": ['content_org:another_org'],
            "expected_response": 403,
            "expected_log": JwtOrgFilterErrorLog
        },
        {
            "filters": [],
            "expected_response": 403,
            "expected_log": JwtOrgFilterErrorLog
        },
    )
    @ddt.unpack
    def test_jwt_org_filters(self, filters, expected_response, expected_log):
        self._assert_jwt_restricted_case(
            scopes=['required_scope'],
            filters=filters,
            is_user_in_url=False,
            expected_response=expected_response,
            expected_log=expected_log,
        )

    @ddt.data(
        # valid user filters
        {
            "user_filters": ['user:me'], "is_user_in_url": True,
            "expected_response": 200,
            "expected_log": JwtIsRestrictedDebugLog,
        },

        # invalid user filters
        {
            "user_filters": ['user:me'], "is_user_in_url": False,
            "expected_response": 403,
            "expected_log": JwtUserFilterErrorLog,
        },
        {
            "user_filters": ['user:another'], "is_user_in_url": True,
            "expected_response": 403,
            "expected_log": JwtUserFilterErrorLog,
        },
    )
    @ddt.unpack
    def test_jwt_user_filters(self, user_filters, is_user_in_url, expected_response, expected_log):
        self._assert_jwt_restricted_case(
            scopes=['required_scope'],
            filters=['content_org:some_org'] + user_filters,
            is_user_in_url=is_user_in_url,
            expected_response=expected_response,
            expected_log=expected_log,
        )

    def _assert_jwt_restricted_case(self, scopes, filters, is_user_in_url, expected_response, expected_log):
        with patch('edx_rest_framework_extensions.permissions.log') as mock_log:
            user = self._create_user()

            auth_header = self._create_jwt_header(user, is_restricted=True, scopes=scopes, filters=filters)
            request = self._create_request(
                username_in_url=user.username if is_user_in_url else None,
                auth_header=auth_header,
            )

            response = self.SomeClassView().dispatch(request, course_id='some_org/course/run')
            self.assertEqual(response.status_code, expected_response)
            self._assert_log(mock_log, expected_log)



"""
Unit tests for jwt authentication middlewares.
"""
from itertools import product
from unittest.mock import ANY, patch

import ddt
from django.http.cookie import SimpleCookie
from django.test import Client, RequestFactory, TestCase, override_settings
from django.urls import re_path as url_pattern
from django.utils.deprecation import MiddlewareMixin
from edx_django_utils.cache import RequestCache
from rest_framework.authentication import SessionAuthentication
from rest_framework.decorators import api_view
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.viewsets import ViewSet
from rest_framework_jwt.authentication import JSONWebTokenAuthentication

from edx_rest_framework_extensions.auth.jwt.constants import USE_JWT_COOKIE_HEADER
from edx_rest_framework_extensions.auth.jwt.cookies import (
    jwt_cookie_header_payload_name,
    jwt_cookie_name,
    jwt_cookie_signature_name,
)
from edx_rest_framework_extensions.auth.jwt.middleware import (
    EnsureJWTAuthSettingsMiddleware,
    JwtAuthCookieMiddleware,
    JwtRedirectToLoginIfUnauthenticatedMiddleware,
)
from edx_rest_framework_extensions.config import ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE
from edx_rest_framework_extensions.permissions import (
    IsStaff,
    IsSuperuser,
    JwtHasContentOrgFilterForRequestedCourse,
    LoginRedirectIfUnauthenticated,
    NotJwtRestrictedApplication,
)
from edx_rest_framework_extensions.tests.factories import UserFactory


class SomeIncludedPermissionClass:
    pass


class SomeJwtAuthenticationSubclass(JSONWebTokenAuthentication):
    pass


def some_auth_decorator(include_jwt_auth, include_required_perm):
    def _decorator(f):
        f.permission_classes = (SomeIncludedPermissionClass,)
        f.authentication_classes = (SessionAuthentication,)
        if include_jwt_auth:
            f.authentication_classes += (SomeJwtAuthenticationSubclass,)
        if include_required_perm:
            f.permission_classes += (NotJwtRestrictedApplication,)
        return f
    return _decorator


@ddt.ddt
class TestEnsureJWTAuthSettingsMiddleware(TestCase):
    def setUp(self):
        super().setUp()
        self.request = RequestFactory().get('/')
        self.middleware = EnsureJWTAuthSettingsMiddleware()  # pylint: disable=no-value-for-parameter

    def _assert_included(self, item, iterator, should_be_included):
        if should_be_included:
            self.assertIn(item, iterator)
        else:
            self.assertNotIn(item, iterator)

    @ddt.data(
        *product(
            ('view_set', 'class_view', 'function_view'),
            (True, False),
            (True, False),
        )
    )
    @ddt.unpack
    def test_api_views(self, view_type, include_jwt_auth, include_required_perm):
        @some_auth_decorator(include_jwt_auth, include_required_perm)
        class SomeClassView(APIView):
            pass

        @some_auth_decorator(include_jwt_auth, include_required_perm)
        class SomeClassViewSet(ViewSet):
            pass

        @api_view(["GET"])
        @some_auth_decorator(include_jwt_auth, include_required_perm)
        def some_function_view(request):  # pylint: disable=unused-argument
            pass

        views = {
            "class_view": SomeClassView,
            "view_set": SomeClassViewSet.as_view({'get': 'list'}),
            "function_view": some_function_view
        }
        view_classes = {
            "class_view": SomeClassView,
            "view_set": views['view_set'].cls,  # pylint: disable=no-member
            "function_view": views['function_view'].view_class,
        }
        view = views[view_type]
        view_class = view_classes[view_type]

        # verify pre-conditions
        self._assert_included(
            SomeJwtAuthenticationSubclass,
            view_class.authentication_classes,
            should_be_included=include_jwt_auth,
        )

        with patch('edx_rest_framework_extensions.auth.jwt.middleware.log.info') as mock_info:
            self.assertIsNone(
                self.middleware.process_view(self.request, view, None, None)
            )
            self.assertEqual(mock_info.called, include_jwt_auth and not include_required_perm)

        # verify post-conditions

        # verify permission class updates
        self._assert_included(
            NotJwtRestrictedApplication,
            view_class.permission_classes,
            should_be_included=include_required_perm or include_jwt_auth,
        )

    def test_simple_view(self):
        """
        Verify middleware works for views that don't have an api_view decorator.
        """
        def some_simple_view(request):  # pylint: disable=unused-argument
            pass

        self.assertIsNone(
            self.middleware.process_view(self.request, some_simple_view, None, None)
        )

    def test_conditional_permissions_drf(self):
        """
        Make sure we handle conditional permissions composed using the inbuilt-support in DRF>=3.9
        """
        class HasCondPermView(APIView):
            authentication_classes = (SomeJwtAuthenticationSubclass,)
            original_permission_classes = (
                JwtHasContentOrgFilterForRequestedCourse & NotJwtRestrictedApplication,
                IsSuperuser | IsStaff,
            )
            permission_classes = original_permission_classes

        class HasNegatedCondPermView(APIView):
            authentication_classes = (SomeJwtAuthenticationSubclass,)
            original_permission_classes = (
                ~NotJwtRestrictedApplication & JwtHasContentOrgFilterForRequestedCourse,
                IsSuperuser | IsStaff,
            )
            permission_classes = original_permission_classes

        class HasNoCondPermView(APIView):
            authentication_classes = (SomeJwtAuthenticationSubclass,)
            original_permission_classes = (
                JwtHasContentOrgFilterForRequestedCourse,
                IsSuperuser | IsStaff,
            )
            permission_classes = original_permission_classes

        # NotJwtRestrictedApplication exists (it's nested in a conditional), so the middleware
        # shouldn't modify this class.
        self.middleware.process_view(self.request, HasCondPermView, None, None)

        # Note: ConditionalPermissions don't implement __eq__
        self.assertIs(
            HasCondPermView.original_permission_classes,
            HasCondPermView.permission_classes
        )

        # NotJwtRestrictedApplication exists (it's nested in a conditional), so the middleware
        # shouldn't modify this class.
        self.middleware.process_view(self.request, HasNegatedCondPermView, None, None)

        # Note: ConditionalPermissions don't implement __eq__
        self.assertIs(
            HasNegatedCondPermView.original_permission_classes,
            HasNegatedCondPermView.permission_classes
        )

        # NotJwtRestrictedApplication does not exist anywhere, so it should be appended
        self.middleware.process_view(self.request, HasNoCondPermView, None, None)

        # Note: ConditionalPermissions don't implement __eq__
        self.assertIsNot(
            HasNoCondPermView.original_permission_classes,
            HasNoCondPermView.permission_classes
        )
        self.assertIn(NotJwtRestrictedApplication, HasNoCondPermView.permission_classes)


class MockJwtAuthentication(JSONWebTokenAuthentication):
    """
    Authenticates a user if the reconstituted jwt cookie contains the expected value.

    The reconstituted value would only be available if JwtAuthCookieMiddleware can
    correctly reconstitute it.
    """
    def authenticate(self, request):
        if request.COOKIES.get(jwt_cookie_name(), None) != 'header.payload.signature':
            return None

        # authenticate was failing on POST calls because it previously wasn't passing the
        # supported parsers into the new Request object. This retrieval of POST data
        # simulates the CSRF POST data checks in the non-test authenticate method. This
        # line lets us verify that the parsers are being correctly passed to the request
        request.POST.get('csrfmiddlewaretoken', '')

        user = UserFactory()
        return (user, None)


class MockUnauthenticatedView(APIView):
    def get(self, request):  # pylint: disable=unused-argument
        return Response({'success': True})


class MockJwtAuthenticationView(APIView):
    authentication_classes = (MockJwtAuthentication,)

    def get(self, request):  # pylint: disable=unused-argument
        return Response({'success': True})

    def post(self, request):  # pylint: disable=unused-argument
        return Response({'success': True})


class LoginRedirectIfUnauthenticatedView(MockJwtAuthenticationView):
    permission_classes = (LoginRedirectIfUnauthenticated,)


class IsAuthenticatedAndLoginRedirectIfUnauthenticatedView(MockJwtAuthenticationView):
    """ If unnecessary IsAuthenticated is added, LoginRedirectIfUnauthenticated should still work. """
    permission_classes = (IsAuthenticated, LoginRedirectIfUnauthenticated,)


class IsAuthenticatedView(MockJwtAuthenticationView):
    permission_classes = (IsAuthenticated,)


class NoPermissionsRequiredView(MockJwtAuthenticationView):
    pass


urlpatterns = [
    url_pattern(
        r'^loginredirectifunauthenticated/$',
        LoginRedirectIfUnauthenticatedView.as_view(),
    ),
    url_pattern(
        r'^isauthenticatedandloginredirect/$',
        IsAuthenticatedAndLoginRedirectIfUnauthenticatedView.as_view()),
    url_pattern(
        r'^isauthenticated/$',
        IsAuthenticatedView.as_view(),
    ),
    url_pattern(
        r'^nopermissionsrequired/$',
        NoPermissionsRequiredView.as_view(),
    ),
    url_pattern(
        r'^unauthenticated/$',
        MockUnauthenticatedView.as_view(),
    ),
]


class OverriddenJwtRedirectToLoginIfUnauthenticatedMiddleware(JwtRedirectToLoginIfUnauthenticatedMiddleware):
    def get_login_url(self, request):
        return '/overridden/login/'

    def is_jwt_auth_enabled_with_login_required(self, request, view_func):
        # force disable jwt cookie reconstitution in all cases
        return False


@ddt.ddt
class TestJwtRedirectToLoginIfUnauthenticatedMiddleware(TestCase):
    """
    Tests integration of JwtRedirectToLoginIfUnauthenticatedMiddleware with JwtAuthCookieMiddleware.

    Note: using an integration test over a unit test, because initial integration was broken.
    """

    def setUp(self):
        super().setUp()
        RequestCache.clear_all_namespaces()
        self.client = Client()

    @ddt.data(
        ('/loginredirectifunauthenticated/', False, 302),
        ('/loginredirectifunauthenticated/', True, 200),
        ('/isauthenticatedandloginredirect/', False, 302),
        ('/isauthenticatedandloginredirect/', True, 200),
        ('/isauthenticated/', False, 401),
        ('/isauthenticated/', True, 401),
        ('/nopermissionsrequired/', False, 200),
        ('/nopermissionsrequired/', True, 200),
    )
    @ddt.unpack
    @override_settings(
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_middleware',
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'edx_rest_framework_extensions.auth.jwt.middleware.JwtRedirectToLoginIfUnauthenticatedMiddleware',
            'edx_rest_framework_extensions.auth.jwt.middleware.JwtAuthCookieMiddleware',
        ),
        LOGIN_URL='/test/login/',
    )
    def test_login_required_middleware(self, url, has_jwt_cookies, expected_status):
        if has_jwt_cookies:
            self.client.cookies = _get_test_cookie()
        response = self.client.get(url)
        self.assertEqual(expected_status, response.status_code)
        if response.status_code == 302:
            self.assertEqual('/test/login/?next=' + url, response.url)

    @ddt.data(
        ('/loginredirectifunauthenticated/', False, 302),
        ('/loginredirectifunauthenticated/', True, 302),
        ('/isauthenticatedandloginredirect/', False, 302),
        ('/isauthenticatedandloginredirect/', True, 302),
        ('/isauthenticated/', False, 401),
        ('/isauthenticated/', True, 401),
        ('/nopermissionsrequired/', False, 200),
        ('/nopermissionsrequired/', True, 200),
    )
    @ddt.unpack
    @override_settings(
        ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_middleware',
        MIDDLEWARE=(
            'django.contrib.sessions.middleware.SessionMiddleware',
            'edx_rest_framework_extensions.auth.jwt.tests.test_middleware.OverriddenJwtRedirectToLoginIfUnauthenticatedMiddleware',  # noqa E501 line too long
            'edx_rest_framework_extensions.auth.jwt.middleware.JwtAuthCookieMiddleware',
        ),
        LOGIN_URL='/test/login/',
    )
    def test_login_required_overridden_middleware(self, url, has_jwt_cookies, expected_status):
        if has_jwt_cookies:
            self.client.cookies = _get_test_cookie()
        response = self.client.get(url)
        self.assertEqual(expected_status, response.status_code)
        if response.status_code == 302:
            self.assertEqual('/overridden/login/?next=' + url, response.url)


class CheckRequestUserForJwtAuthMiddleware(MiddlewareMixin):
    """
    This test middleware can be used to confirm that our Jwt Authentication related middleware correctly
    set the request.user.
    """
    def process_view(self, request, view_func, view_args, view_kwargs):  # pylint: disable=unused-argument
        assert request.user.is_authenticated, 'Request.user was expected to be authenticated.'


class CheckRequestUserAnonymousForJwtAuthMiddleware(MiddlewareMixin):
    """
    This test middleware can be used to confirm that when Jwt Authentication related middleware does not set
    the user (e.g. a failed cookie).
    """
    def process_view(self, request, view_func, view_args, view_kwargs):  # pylint: disable=unused-argument
        assert not request.user.is_authenticated, 'Request.user was expected to be anonymous.'


@ddt.ddt
class TestJwtAuthCookieMiddleware(TestCase):
    def setUp(self):
        super().setUp()
        self.request = RequestFactory().get('/')
        self.request.session = 'mock session'
        self.middleware = JwtAuthCookieMiddleware()  # pylint: disable=no-value-for-parameter

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_do_not_use_jwt_cookies(self, mock_set_custom_attribute):
        self.middleware.process_view(self.request, None, None, None)
        self.assertIsNone(self.request.COOKIES.get(jwt_cookie_name()))
        mock_set_custom_attribute.assert_called_once_with('request_jwt_cookie', 'not-requested')

    @ddt.data(
        (jwt_cookie_header_payload_name(), jwt_cookie_signature_name()),
        (jwt_cookie_signature_name(), jwt_cookie_header_payload_name()),
    )
    @ddt.unpack
    @patch('edx_rest_framework_extensions.auth.jwt.middleware.log')
    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_missing_cookies(
            self, set_cookie_name, missing_cookie_name, mock_set_custom_attribute, mock_log
    ):
        self.request.META[USE_JWT_COOKIE_HEADER] = 'true'
        self.request.COOKIES[set_cookie_name] = 'test'
        self.middleware.process_view(self.request, None, None, None)
        self.assertIsNone(self.request.COOKIES.get(jwt_cookie_name()))
        mock_log.warning.assert_called_once_with(
            '%s cookie is missing. JWT auth cookies will not be reconstituted.' %
            missing_cookie_name
        )
        mock_set_custom_attribute.assert_called_once_with(
            'request_jwt_cookie', f'missing-{missing_cookie_name}'
        )

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_no_cookies(self, mock_set_custom_attribute):
        self.request.META[USE_JWT_COOKIE_HEADER] = 'true'
        self.middleware.process_view(self.request, None, None, None)
        self.assertIsNone(self.request.COOKIES.get(jwt_cookie_name()))
        mock_set_custom_attribute.assert_called_once_with('request_jwt_cookie', 'missing-both')

    @patch('edx_django_utils.monitoring.set_custom_attribute')
    def test_success(self, mock_set_custom_attribute):
        self.request.META[USE_JWT_COOKIE_HEADER] = 'true'
        self.request.COOKIES[jwt_cookie_header_payload_name()] = 'header.payload'
        self.request.COOKIES[jwt_cookie_signature_name()] = 'signature'
        self.middleware.process_view(self.request, None, None, None)
        self.assertEqual(self.request.COOKIES[jwt_cookie_name()], 'header.payload.signature')
        mock_set_custom_attribute.assert_called_once_with('request_jwt_cookie', 'success')

    _LOG_WARN_AUTHENTICATION_FAILED = 0
    _LOG_WARN_MISSING_JWT_AUTHENTICATION_CLASS = 1

    @patch('edx_rest_framework_extensions.auth.jwt.middleware.log')
    @ddt.data(
        ('/nopermissionsrequired/', True, True, True, None, False),
        ('/nopermissionsrequired/', True, False, False, None, False),
        ('/nopermissionsrequired/', False, False, True, _LOG_WARN_AUTHENTICATION_FAILED, False),
        ('/nopermissionsrequired/', False, False, False, None, False),
        ('/nopermissionsrequired/', True, True, True, None, True),
        ('/unauthenticated/', True, False, True, _LOG_WARN_MISSING_JWT_AUTHENTICATION_CLASS, False),
        ('/unauthenticated/', True, False, True, None, False),
    )
    @ddt.unpack
    def test_set_request_user_with_use_jwt_cookie(
            self, url, is_cookie_valid, is_request_user_set, is_toggle_enabled, log_warning,
            send_post, mock_log,
    ):
        header = {USE_JWT_COOKIE_HEADER: 'true'}
        self.client.cookies = _get_test_cookie(is_cookie_valid=is_cookie_valid)
        check_user_middleware_assertion_class = (
            'CheckRequestUserForJwtAuthMiddleware'
            if is_request_user_set else
            'CheckRequestUserAnonymousForJwtAuthMiddleware'
        )
        with override_settings(
            ROOT_URLCONF='edx_rest_framework_extensions.auth.jwt.tests.test_middleware',
            MIDDLEWARE=(
                    'django.contrib.sessions.middleware.SessionMiddleware',
                    'edx_rest_framework_extensions.auth.jwt.middleware.JwtAuthCookieMiddleware',
                    'django.contrib.auth.middleware.AuthenticationMiddleware',
                    'edx_rest_framework_extensions.auth.jwt.tests.test_middleware.{}'.format(
                        check_user_middleware_assertion_class
                    ),
            ),
            EDX_DRF_EXTENSIONS={
                ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE: is_toggle_enabled,
            }
        ):
            if send_post:
                response = self.client.post(url, {}, content_type="application/json", **header)
            else:
                response = self.client.get(url, **header)
            self.assertEqual(200, response.status_code)

            if log_warning == self._LOG_WARN_AUTHENTICATION_FAILED:
                mock_log.warning.assert_called_once_with('Jwt Authentication failed and request.user could not be set.')
            elif log_warning == self._LOG_WARN_MISSING_JWT_AUTHENTICATION_CLASS:
                mock_log.warning.assert_called_once_with(
                    'Jwt Authentication expected, but view %s is not using a JwtAuthentication class.', ANY
                )
            else:
                mock_log.warn.assert_not_called()


def _get_test_cookie(is_cookie_valid=True):
    header_payload_value = 'header.payload' if is_cookie_valid else 'header.payload.invalid'
    return SimpleCookie({
        jwt_cookie_header_payload_name(): header_payload_value,
        jwt_cookie_signature_name(): 'signature',
    })



"""
JWT decoder utility functions.

In most of this module, "decode" refers to both verifying and unpacking a JWT,
as a unified operation. (Reading the contents of an unverified JWT would be
a security risk in the general case.)
"""
import logging
import sys

import jwt
from django.conf import settings
from edx_django_utils.monitoring import set_custom_attribute
from jwkest.jwk import KEYS
from jwkest.jws import JWS
from rest_framework_jwt.settings import api_settings
from semantic_version import Version

from edx_rest_framework_extensions.settings import get_first_jwt_issuer, get_jwt_issuers


logger = logging.getLogger(__name__)


class JwtTokenVersion:
    default_latest_supported = '1.2.0'

    starting_version = '1.0.0'
    added_version = '1.1.0'


def jwt_decode_handler(token, decode_symmetric_token=True):
    """
    Decodes (and verifies) a JSON Web Token (JWT).

    Notes:
        * Requires "exp" and "iat" claims to be present in the token's payload.
        * Aids debugging by logging InvalidTokenError log entries when decoding fails.
        * Setting for JWT_DECODE_HANDLER expects a single argument, token. The argument decode_symmetric_token
          is for internal use only.

    Examples:
        Use with `djangorestframework-jwt <https://getblimp.github.io/django-rest-framework-jwt/>`_, by changing
        your Django settings:

        .. code-block:: python

            JWT_AUTH = {
                'JWT_DECODE_HANDLER': 'edx_rest_framework_extensions.auth.jwt.decoder.jwt_decode_handler',
                'JWT_ISSUER': 'https://the.jwt.issuer',
                'JWT_SECRET_KEY': 'the-jwt-secret-key',  (defaults to settings.SECRET_KEY)
                'JWT_AUDIENCE': 'the-jwt-audience',
                'JWT_PUBLIC_SIGNING_JWK_SET': 'the-jwk-set-of-public-signing-keys',
            }

    Warning:
        Do **not** use this method internally.  Only use it in ``JWT_DECODE_HANDLER`` like the above example.
        Internally, use ``configured_jwt_decode_handler`` which respects the ``JWT_DECODE_HANDLER`` setting.

    Args:
        token (str): JWT to be decoded.
        decode_symmetric_token (bool): Whether to decode symmetric tokens or not. Pass False for asymmetric tokens only

    Returns:
        dict: Decoded JWT payload.

    Raises:
        MissingRequiredClaimError: Either the exp or iat claims is missing from the JWT payload.
        InvalidTokenError: Decoding fails.
    """
    jwt_issuer = get_first_jwt_issuer()
    _verify_jwt_signature(token, jwt_issuer, decode_symmetric_token=decode_symmetric_token)
    decoded_token = _decode_and_verify_token(token, jwt_issuer)
    return _set_token_defaults(decoded_token)


def configured_jwt_decode_handler(token):
    """
    Calls the ``jwt_decode_handler`` configured in the ``JWT_DECODE_HANDLER`` setting.
    """
    api_setting_jwt_decode_handler = api_settings.JWT_DECODE_HANDLER
    return api_setting_jwt_decode_handler(token)


def get_asymmetric_only_jwt_decode_handler(token):
    """
    Returns a jwt_decode_handler that will only validate asymmetrically signed JWTs.

    WARNING: This will only work with a service that is configured to use the
       jwt_decode_handler from this library. This can be used to decode an
       already decoded JWT, to ensure it is asymmetrically signed. This check
       can go away once the DEPR for symmetrically signed JWTs is complete:
       https://github.com/openedx/public-engineering/issues/83
    """
    return jwt_decode_handler(token, decode_symmetric_token=False)


def decode_jwt_scopes(token):
    """
    Decode the JWT and return the scopes claim.
    """
    return configured_jwt_decode_handler(token).get('scopes', [])


def decode_jwt_is_restricted(token):
    """
    Decode the JWT and return the is_restricted claim.
    """
    return configured_jwt_decode_handler(token).get('is_restricted', False)


def decode_jwt_filters(token):
    """
    Decode the JWT, parse the filters claim, and return a
    list of (filter_type, filter_value) tuples.
    """
    return [
        jwt_filter.split(':')
        for jwt_filter in configured_jwt_decode_handler(token).get('filters', [])
    ]


def _set_token_defaults(token):
    """
    Returns an updated token that includes default values for
    fields that were introduced since the token was created
    by checking its version number.
    """
    def _verify_version(jwt_version):
        supported_version = Version(
            settings.JWT_AUTH.get('JWT_SUPPORTED_VERSION', JwtTokenVersion.default_latest_supported)
        )
        if jwt_version.major > supported_version.major:
            logger.info('Token decode failed due to unsupported JWT version number [%s]', str(jwt_version))
            raise jwt.InvalidTokenError('JWT version number [%s] is unsupported' % str(jwt_version))

    def _get_and_set_version(token):
        """
        Tokens didn't always contain a version number so we
        default to a nominal starting number.
        """
        if 'version' not in token:
            token['version'] = str(JwtTokenVersion.starting_version)

        return Version(token['version'])

    def _set_is_restricted(token):
        """
        We can safely default to False since all "restricted" tokens
        created prior to the addition of the `is_restricted` flag were
        always created as expired tokens. Expired tokens would not
        validate and so would not get this far into the decoding process.
        # TODO: ARCH-166
        """
        if 'is_restricted' not in token:
            token['is_restricted'] = False

    def _set_filters(token):
        """
        We can safely default to an empty list of filters since
        previously created tokens were either "restricted" (always
        expired) or had full access.
        # TODO: ARCH-166
        """
        if 'filters' not in token:
            token['filters'] = []

    token_version = _get_and_set_version(token)
    _verify_version(token_version)
    _set_is_restricted(token)
    _set_filters(token)
    return token


def _verify_jwt_signature(token, jwt_issuer, decode_symmetric_token):
    """
    Verifies the JWT signature. Raises InvalidTokenError in the event of an error.

    Arguments:
        token (str): JWT to be decoded.
        jwt_issuer (dict): A dict of JWT issuer related settings, containing the symmetric key.
        decode_symmetric_token (bool): Whether to decode symmetric tokens or not. Pass False for asymmetric tokens only
    """
    # .. custom_attribute_name: jwt_auth_check_symmetric_key
    # .. custom_attribute_description: True if symmetric keys will also be used for checking
    #   the JWT signature, and False if only asymmetric keys will be used.
    set_custom_attribute('jwt_auth_check_symmetric_key', decode_symmetric_token)

    # For observability purposes, we will first try asymmetric keys only to verify
    #   that we no longer need the symmetric key. However, if this fails, we will
    #   continue on to the original code path and try all keys (including symmetric)
    #   and add monitoring to let us know. This is meant to be temporary, until we
    #   can fully retire code paths for symmetric keys, as part of
    #   DEPR: Symmetric JWTs: https://github.com/openedx/public-engineering/issues/83

    # Use add_symmetric_keys=False to only include asymmetric keys at first
    key_set = _get_signing_jwk_key_set(jwt_issuer, add_symmetric_keys=False)
    # .. custom_attribute_name: jwt_auth_verify_asymmetric_keys_count
    # .. custom_attribute_description: Number of JWT verification keys in use for this
    #   verification. Should be same as number of asymmetric public keys. This is
    #   intended to aid in key rotations; once the average count stabilizes at a
    #   higher number after adding a public key, it should be safe to change the secret key.
    set_custom_attribute('jwt_auth_verify_asymmetric_keys_count', len(key_set))

    try:
        _ = JWS().verify_compact(token, key_set)
        # .. custom_attribute_name: jwt_auth_asymmetric_verified
        # .. custom_attribute_description: Whether the JWT was successfully verified
        #   using an asymmetric key.
        set_custom_attribute('jwt_auth_asymmetric_verified', True)
        return
    except Exception:  # pylint: disable=broad-except
        # Continue to the old code path of trying all keys
        pass

    # The following is the original code that includes both the symmetric and asymmetric keys
    #   as requested with the decode_symmetric_token argument. Note that the check against
    #   the asymmetric keys here is redundant and unnecessary, but this code is temporary and
    #   will be simplified once symmetric keys have been fully retired.

    key_set = _get_signing_jwk_key_set(jwt_issuer, add_symmetric_keys=decode_symmetric_token)
    # .. custom_attribute_name: jwt_auth_verify_all_keys_count
    # .. custom_attribute_description: Number of JWT verification keys in use for this
    #   verification. Should be same as number of asymmetric public keys, plus one if
    #   a symmetric key secret is set. This is intended to aid in key rotations; once
    #   the average count stabilizes at a higher number after adding a public key, it
    #   should be safe to change the secret key.
    set_custom_attribute('jwt_auth_verify_all_keys_count', len(key_set))

    try:
        _ = JWS().verify_compact(token, key_set)
        # .. custom_attribute_name: jwt_auth_symmetric_verified
        # .. custom_attribute_description: Whether the JWT was successfully verified
        #   using a symmetric key.
        # Note: Rather than using a single custom attribute like ``jwt_auth_verified``
        #   with values of 'symmetric' or 'asymmetric', we use two separate custom
        #   attribute names (e.g. jwt_auth_symmetric_verified and jwt_auth_asymmetric_verified),
        #   so that if each of these were set separately in the same request, they
        #   wouldn't clobber each other.
        set_custom_attribute('jwt_auth_symmetric_verified', True)
        return
    except Exception as token_error:
        # .. custom_attribute_name: jwt_auth_verification_failed
        # .. custom_attribute_description: True if the JWT token verification failed.
        set_custom_attribute('jwt_auth_verification_failed', True)
        logger.exception('Token verification failed.')
        exc_info = sys.exc_info()
        raise jwt.InvalidTokenError(exc_info[2]) from token_error


def _decode_and_verify_token(token, jwt_issuer):
    """
    Part of the verification implementation; must not be used in isolation,
    as the signature is actually checked in a different function.
    """
    options = {
        'require': ["exp", "iat"],

        'verify_exp': api_settings.JWT_VERIFY_EXPIRATION,
        'verify_aud': settings.JWT_AUTH.get('JWT_VERIFY_AUDIENCE', True),
        'verify_iss': False,  # TODO (ARCH-204): manually verify until issuer is configured correctly.
        'verify_signature': False,  # Verified with JWS already
    }

    decoded_token = jwt.decode(
        token,
        jwt_issuer['SECRET_KEY'],
        options=options,
        leeway=api_settings.JWT_LEEWAY,
        audience=jwt_issuer['AUDIENCE'],
        issuer=jwt_issuer['ISSUER'],
        algorithms=[api_settings.JWT_ALGORITHM],
    )

    # TODO (ARCH-204): verify issuer manually until it is properly configured.
    token_issuer = decoded_token.get('iss')
    # .. custom_attribute_name: jwt_auth_issuer
    # .. custom_attribute_description: Value set to the JWT auth issuer.
    set_custom_attribute('jwt_auth_issuer', token_issuer)
    issuer_matched = any(issuer['ISSUER'] == token_issuer for issuer in get_jwt_issuers())
    if token_issuer == jwt_issuer['ISSUER']:
        # .. custom_attribute_name: jwt_auth_issuer_verification
        # .. custom_attribute_description: Depending on issuer verification, the value will
        #   be one of: matches-first-issuer, matches-later-issuer, or no-match.
        set_custom_attribute('jwt_auth_issuer_verification', 'matches-first-issuer')
    elif issuer_matched:
        set_custom_attribute('jwt_auth_issuer_verification', 'matches-later-issuer')
    else:
        set_custom_attribute('jwt_auth_issuer_verification', 'no-match')
        logger.info('Token decode failed due to mismatched issuer [%s]', token_issuer)
        raise jwt.InvalidTokenError('%s is not a valid issuer.' % token_issuer)

    return decoded_token


def _get_signing_jwk_key_set(jwt_issuer, add_symmetric_keys=True):
    """
    Returns a JWK Keyset containing all active keys that are configured
    for verifying signatures.
    """
    key_set = KEYS()

    # asymmetric keys
    signing_jwk_set = settings.JWT_AUTH.get('JWT_PUBLIC_SIGNING_JWK_SET')
    if signing_jwk_set:
        key_set.load_jwks(signing_jwk_set)

    if add_symmetric_keys:
        # symmetric key
        key_set.add({'key': jwt_issuer['SECRET_KEY'], 'kty': 'oct'})

    return key_set



"""
NOTE: Support for multiple JWT_ISSUERS is deprecated.  Instead, Asymmetric JWTs
make this simpler by using JWK keysets to list all available public keys.

Settings for edx-drf-extensions are all namespaced in the EDX_DRF_EXTENSIONS setting.
For example your project's `settings.py` file might look like this:

EDX_DRF_EXTENSIONS = {
    'OAUTH2_ACCESS_TOKEN_URL': 'https://example.com/oauth2/access_token'
}
"""
import logging
import warnings

from django.conf import settings
from rest_framework_jwt.settings import api_settings

from edx_rest_framework_extensions.config import ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE


logger = logging.getLogger(__name__)


DEFAULT_SETTINGS = {
    'OAUTH2_USER_INFO_URL': None,

    # Map JWT claims to user attributes.
    'JWT_PAYLOAD_USER_ATTRIBUTE_MAPPING': {
        'administrator': 'is_staff',
        'email': 'email',
    },
    'JWT_PAYLOAD_MERGEABLE_USER_ATTRIBUTES': (),
    ENABLE_SET_REQUEST_USER_FOR_JWT_COOKIE: False,
}


def get_setting(name):
    """ Returns the value of the named setting.

        Arguments:
            name (str): Name of the setting to retrieve

        Raises:
            KeyError: The specified setting does not exist.
    """
    try:
        return getattr(settings, 'EDX_DRF_EXTENSIONS', {})[name]
    except KeyError:
        return DEFAULT_SETTINGS[name]


def _get_current_jwt_issuers():
    """
    Internal helper to retrieve the current set of JWT_ISSUERS from the JWT_AUTH configuration
    Having this allows for easier testing/mocking
    """
    # If we have a 'JWT_ISSUERS' list defined, return it
    return settings.JWT_AUTH.get('JWT_ISSUERS', None)


def _get_deprecated_jwt_issuers():
    """
    Internal helper to retrieve the deprecated set of JWT_ISSUER data from the JWT_AUTH configuration
    Having this allows for easier testing/mocking
    """
    # If JWT_ISSUERS is not defined, attempt to return the deprecated settings.
    warnings.warn(
        "'JWT_ISSUERS' list not defined, checking for deprecated settings.",
        DeprecationWarning
    )

    return [
        {
            'ISSUER': api_settings.JWT_ISSUER,
            'SECRET_KEY': api_settings.JWT_SECRET_KEY,
            'AUDIENCE': api_settings.JWT_AUDIENCE
        }
    ]


def get_jwt_issuers():
    """
    Retrieves the JWT_ISSUERS list from system configuraiton.  If no list is defined in JWT_AUTH/JWT_ISSUERS
    an attempt is made to instead return the deprecated JWT configuration settings.
    """
    # If we have a 'JWT_ISSUERS' list defined, return it
    jwt_issuers = _get_current_jwt_issuers()
    if jwt_issuers:
        return jwt_issuers
    # If we do not, return the deprecated configuration
    return _get_deprecated_jwt_issuers()


def get_first_jwt_issuer():
    """
    Retrieves the first issuer in the JWT_ISSUERS list.

    As mentioned above, support for multiple JWT_ISSUERS is deprecated. They
    are currently used only to distinguish the "ISSUER" field across sites.
    So in many cases, we just need the first issuer value.
    """
    return get_jwt_issuers()[0]



"""
URL definitions for version 1 of the CSRF API.
"""

from django.urls import re_path

from .views import CsrfTokenView


urlpatterns = [
    re_path(r'^token$', CsrfTokenView.as_view(), name='csrf_token'),
]



""" Tests for the CSRF API """

from django.test.utils import override_settings
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APITestCase


class CsrfTokenTests(APITestCase):
    """ Tests for the CSRF token endpoint. """

    @override_settings(REST_FRAMEWORK={
        'DEFAULT_PERMISSION_CLASSES': (
            # Ensure this default permission does not interfere with the CSRF endpoint.
            'rest_framework.permissions.DjangoModelPermissions',
        ),
    })
    def test_get_token(self):
        """
        Ensure we can get a CSRF token for an anonymous user.
        """
        url = reverse('csrf_token')
        response = self.client.get(url, format='json')
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIn('csrfToken', response.data)
        self.assertIsNotNone(response.data['csrfToken'])



#!/usr/bin/env python
import os
import re

from setuptools import find_packages, setup


def is_requirement(line):
    """
    Return True if the requirement line is a package requirement.

    Returns:
        bool: True if the line is not blank, a comment,
        a URL, or an included file
    """
    # UPDATED VIA SEMGREP - if you need to remove/modify this method remove this line and add a comment specifying why

    return line and line.strip() and not line.startswith(('-r', '#', '-e', 'git+', '-c'))


def load_requirements(*requirements_paths):
    """
    Load all requirements from the specified requirements files.

    Requirements will include any constraints from files specified
    with -c in the requirements files.
    Returns a list of requirement strings.
    """
    # UPDATED VIA SEMGREP - if you need to remove/modify this method remove this line and add a comment specifying why.
    # small change from original SEMGREP update to allow packages with [] in the name

    requirements = {}
    constraint_files = set()

    # groups "my-package-name<=x.y.z,..." into ("my-package-name", "<=x.y.z,...")
    requirement_line_regex = re.compile(r"([a-zA-Z0-9-_.\[\]]+)([<>=][^#\s]+)?")

    def add_version_constraint_or_raise(current_line, current_requirements, add_if_not_present):
        regex_match = requirement_line_regex.match(current_line)
        if regex_match:
            package = regex_match.group(1)
            version_constraints = regex_match.group(2)
            existing_version_constraints = current_requirements.get(package, None)
            # it's fine to add constraints to an unconstrained package, but raise an error if there are already
            # constraints in place
            if existing_version_constraints and existing_version_constraints != version_constraints:
                raise BaseException(f'Multiple constraint definitions found for {package}:'
                                    f' "{existing_version_constraints}" and "{version_constraints}".'
                                    f'Combine constraints into one location with {package}'
                                    f'{existing_version_constraints},{version_constraints}.')
            if add_if_not_present or package in current_requirements:
                current_requirements[package] = version_constraints

    # process .in files and store the path to any constraint files that are pulled in
    for path in requirements_paths:
        with open(path) as reqs:
            for line in reqs:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, True)
                if line and line.startswith('-c') and not line.startswith('-c http'):
                    constraint_files.add(os.path.dirname(path) + '/' + line.split('#')[0].replace('-c', '').strip())

    # process constraint files and add any new constraints found to existing requirements
    for constraint_file in constraint_files:
        with open(constraint_file) as reader:
            for line in reader:
                if is_requirement(line):
                    add_version_constraint_or_raise(line, requirements, False)

    # process back into list of pkg><=constraints strings
    constrained_requirements = [f'{pkg}{version or ""}' for (pkg, version) in sorted(requirements.items())]
    return constrained_requirements


def get_version(*file_paths):
    """
    Extract the version string from the file at the given relative path fragments.
    """
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    with open(filename, encoding='utf-8') as opened_file:
        version_file = opened_file.read()
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                                  version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


VERSION = get_version("edx_rest_framework_extensions", "__init__.py")

README = open(os.path.join(os.path.dirname(__file__), 'README.rst'), encoding="utf8").read()
CHANGELOG = open(os.path.join(os.path.dirname(__file__), 'CHANGELOG.rst'), encoding="utf8").read()


setup(
    name='edx-drf-extensions',
    version=VERSION,
    description='edX extensions of Django REST Framework',
    long_description=README + '\n\n' + CHANGELOG,
    author='edX',
    author_email='oscm@edx.org',
    url='https://github.com/openedx/edx-drf-extensions',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Framework :: Django',
        'Framework :: Django :: 3.2',
    ],
    packages=find_packages(exclude=["tests"]),
    install_requires=load_requirements('requirements/base.in'),
    tests_require=load_requirements('requirements/test.in'),
)



""" Bearer Authentication class. """

import logging

import requests
from django.contrib.auth import get_user_model
from edx_django_utils.monitoring import set_custom_attribute
from rest_framework import exceptions
from rest_framework.authentication import BaseAuthentication, get_authorization_header

from edx_rest_framework_extensions.exceptions import UserInfoRetrievalFailed
from edx_rest_framework_extensions.settings import get_setting


logger = logging.getLogger(__name__)


class BearerAuthentication(BaseAuthentication):
    """
    Simple token based authentication.

    This authentication class is useful for authenticating an OAuth2 access token against a remote
    authentication provider. Clients should authenticate by passing the token key in the "Authorization" HTTP header,
    prepended with the string `"Bearer "`.

    This class relies on the OAUTH2_USER_INFO_URL being set to the value of an endpoint on the OAuth provider, that
    returns a JSON object with information about the user. See ``process_user_info_response`` for the expected format
    of this object. This data will be used to get, or create, a ``User``. Additionally, it is assumed that a successful
    response from this endpoint (authenticated with the provided access token) implies the access token is valid.

    Example Header:
        Authorization: Bearer 401f7ac837da42b97f613d789819ff93537bee6a
    """

    def get_user_info_url(self):
        """ Returns the URL, hosted by the OAuth2 provider, from which user information can be pulled. """
        return get_setting('OAUTH2_USER_INFO_URL')

    def authenticate(self, request):
        set_custom_attribute("BearerAuthentication", "Failed")  # default value
        if not self.get_user_info_url():
            logger.warning('The setting OAUTH2_USER_INFO_URL is invalid!')
            set_custom_attribute("BearerAuthentication", "NoURL")
            return None
        set_custom_attribute("BearerAuthentication_user_info_url", self.get_user_info_url())
        auth = get_authorization_header(request).split()

        if not auth or auth[0].lower() != b'bearer':
            set_custom_attribute("BearerAuthentication", "None")
            return None

        if len(auth) == 1:
            raise exceptions.AuthenticationFailed('Invalid token header. No credentials provided.')
        if len(auth) > 2:
            raise exceptions.AuthenticationFailed('Invalid token header. Token string should not contain spaces.')

        output = self.authenticate_credentials(auth[1].decode('utf8'))
        set_custom_attribute("BearerAuthentication", "Success")
        return output

    def authenticate_credentials(self, token):
        """
        Validate the bearer token against the OAuth provider.

        Arguments:
            token (str): Access token to validate

        Returns:
            (tuple): tuple containing:

                user (User): User associated with the access token
                access_token (str): Access token

        Raises:
            AuthenticationFailed: The user is inactive, or retrieval of user info failed.
        """

        try:
            user_info = self.get_user_info(token)
        except UserInfoRetrievalFailed as authentication_error:
            msg = 'Failed to retrieve user info. Unable to authenticate.'
            logger.error(msg)
            raise exceptions.AuthenticationFailed(msg) from authentication_error

        user, __ = get_user_model().objects.get_or_create(username=user_info['username'], defaults=user_info)

        if not user.is_active:
            raise exceptions.AuthenticationFailed('User inactive or deleted.')

        return user, token

    def get_user_info(self, token):
        """
        Retrieves the user info from the OAuth provider.

        Arguments:
            token (str): OAuth2 access token.

        Returns:
            dict

        Raises:
            UserInfoRetrievalFailed: Retrieval of user info from the remote server failed.
        """

        url = self.get_user_info_url()

        try:
            headers = {'Authorization': f'Bearer {token}'}
            response = requests.get(url, headers=headers)  # pylint: disable=missing-timeout
        except requests.RequestException as error:
            logger.exception('Failed to retrieve user info due to a request exception.')
            raise UserInfoRetrievalFailed from error

        if response.status_code == 200:
            return self.process_user_info_response(response.json())
        else:
            msg = 'Failed to retrieve user info. Server [{server}] responded with status [{status}].'.format(
                server=url,
                status=response.status_code
            )
            raise UserInfoRetrievalFailed(msg)

    def process_user_info_response(self, response):
        """
        Process the user info response data.

        By default, this simply maps the edX user info key-values (example below) to Django-friendly names. If your
        provider returns different fields, you should sub-class this class and override this method.

        .. code-block:: python

            {
                "username": "jdoe",
                "email": "jdoe@example.com",
                "first_name": "Jane",
                "last_name": "Doe"
            }

        Arguments:
            response (dict): User info data

        Returns:
            dict
        """
        mapping = (
            ('username', 'preferred_username'),
            ('email', 'email'),
            ('last_name', 'family_name'),
            ('first_name', 'given_name'),
        )

        return {dest: response[source] for dest, source in mapping}

    def authenticate_header(self, request):
        return 'Bearer'



""" Tests paginator methods """

from collections import namedtuple
from unittest import TestCase
from unittest.mock import MagicMock, Mock

import ddt
from django.http import Http404
from django.test import RequestFactory
from rest_framework import serializers

from edx_rest_framework_extensions.paginators import (
    NamespacedPageNumberPagination,
    paginate_search_results,
)


@ddt.ddt
class PaginateSearchResultsTestCase(TestCase):
    """Test cases for paginate_search_results method"""

    def setUp(self):
        super().setUp()

        self.default_size = 6
        self.default_page = 1
        self.search_results = {
            "count": 3,
            "took": 1,
            "results": [
                {
                    '_id': 0,
                    'data': {
                        'pk': 0,
                        'name': 'object 0'
                    }
                },
                {
                    '_id': 1,
                    'data': {
                        'pk': 1,
                        'name': 'object 1'
                    }
                },
                {
                    '_id': 2,
                    'data': {
                        'pk': 2,
                        'name': 'object 2'
                    }
                },
                {
                    '_id': 3,
                    'data': {
                        'pk': 3,
                        'name': 'object 3'
                    }
                },
                {
                    '_id': 4,
                    'data': {
                        'pk': 4,
                        'name': 'object 4'
                    }
                },
                {
                    '_id': 5,
                    'data': {
                        'pk': 5,
                        'name': 'object 5'
                    }
                },
            ]
        }
        self.mock_model = Mock()
        self.mock_model.objects = Mock()
        self.mock_model.objects.filter = Mock()

    @ddt.data(
        (1, 1, True),
        (1, 3, True),
        (1, 5, True),
        (1, 10, False),
        (2, 1, True),
        (2, 3, False),
        (2, 5, False),
    )
    @ddt.unpack
    def test_paginated_results(self, page_number, page_size, has_next):
        """ Test the page returned has the expected db objects and acts
        like a proper page object.
        """
        id_range = get_object_range(page_number, page_size)
        db_objects = [build_mock_object(obj_id) for obj_id in id_range]
        self.mock_model.objects.filter = MagicMock(return_value=db_objects)

        page = paginate_search_results(self.mock_model, self.search_results, page_size, page_number)

        self.mock_model.objects.filter.assert_called_with(pk__in=id_range)
        self.assertEqual(db_objects, page.object_list)
        self.assertTrue(page.number, page_number)
        self.assertEqual(page.has_next(), has_next)

    def test_paginated_results_last_keyword(self):
        """ Test the page returned has the expected db objects and acts
        like a proper page object using 'last' keyword.
        """
        page_number = 2
        page_size = 3
        id_range = get_object_range(page_number, page_size)
        db_objects = [build_mock_object(obj_id) for obj_id in id_range]
        self.mock_model.objects.filter = MagicMock(return_value=db_objects)
        page = paginate_search_results(self.mock_model, self.search_results, page_size, 'last')

        self.mock_model.objects.filter.assert_called_with(pk__in=id_range)
        self.assertEqual(db_objects, page.object_list)
        self.assertTrue(page.number, page_number)
        self.assertFalse(page.has_next())

    @ddt.data(10, -1, 0, 'str')
    def test_invalid_page_number(self, page_num):
        """ Test that a Http404 error is raised with non-integer and out-of-range pages
        """
        with self.assertRaises(Http404):
            paginate_search_results(self.mock_model, self.search_results, self.default_size, page_num)


class NamespacedPaginationTestCase(TestCase):
    """
    Test behavior of `NamespacedPageNumberPagination`
    """

    TestUser = namedtuple('TestUser', ['username', 'email'])

    class TestUserSerializer(serializers.Serializer):  # pylint: disable=abstract-method
        """
        Simple serializer to paginate results from
        """
        username = serializers.CharField()
        email = serializers.CharField()

    expected_data = {
        'results': [
            {'username': 'user_5', 'email': 'user_5@example.com'},
            {'username': 'user_6', 'email': 'user_6@example.com'},
            {'username': 'user_7', 'email': 'user_7@example.com'},
            {'username': 'user_8', 'email': 'user_8@example.com'},
            {'username': 'user_9', 'email': 'user_9@example.com'},
        ],
        'pagination': {
            'next': 'http://testserver/endpoint?page=3&page_size=5',
            'previous': 'http://testserver/endpoint?page_size=5',
            'count': 25,
            'num_pages': 5,
        }
    }

    def setUp(self):
        super().setUp()
        self.paginator = NamespacedPageNumberPagination()
        self.users = [self.TestUser(f'user_{idx}', f'user_{idx}@example.com') for idx in range(25)]
        self.request_factory = RequestFactory()

    def test_basic_pagination(self):
        request = self.request_factory.get('/endpoint', data={'page': 2, 'page_size': 5})
        request.query_params = {'page': 2, 'page_size': 5}
        paged_users = self.paginator.paginate_queryset(self.users, request)
        results = self.TestUserSerializer(paged_users, many=True).data
        self.assertEqual(self.expected_data, self.paginator.get_paginated_response(results).data)


def build_mock_object(obj_id):
    """ Build a mock object with the passed id"""
    mock_object = Mock()
    object_config = {
        'pk': obj_id,
        'name': f"object {obj_id}"
    }
    mock_object.configure_mock(**object_config)
    return mock_object


def get_object_range(page, page_size):
    """ Get the range of expected object ids given a page and page size.
    This will take into account the max_id of the sample data.  Currently 5.
    """
    max_id = 5
    start = min((page - 1) * page_size, max_id)
    end = min(start + page_size, max_id + 1)
    return list(range(start, end))
